import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression

from faststylometry.burrows_delta import calculate_burrows_delta
from faststylometry.corpus import Corpus


def every_item_but_one(l: list, idx: int) -> list:
    """
    Returns every item in the list except for the one at index idx.
    :param l: a list to process.
    :param idx: the index to be excluded from the list.
    :return: the list l minus the item at index idx.
    """
    return [item for i, item in enumerate(l) if i != idx]


def get_calibration_curve(corpus: Corpus) -> tuple:
    """
    Calculates the probability calibration curve of the Burrows' Delta calculation on the train corpus, using a cross-validation technique.
    :param corpus: the corpus for which we desire a probability calibration curve.
    :return: two arrays, an array of ground truths (0: different authors, 1: same author), and of Burrows' delta values. These can be used to calibrate a model such as logistic regression, or to generate a ROC curve.
    """
    ground_truths = []

    delta_values = []

    num_books = len(corpus.authors)

    for i in range(num_books):
        tmp_train_corpus = Corpus(every_item_but_one(corpus.authors, i), every_item_but_one(corpus.books, i),
                                  every_item_but_one(corpus.tokens, i))
        tmp_test_corpus = Corpus(corpus.authors[i:i + 1], corpus.books[i:i + 1], corpus.tokens[i:i + 1])

        true_author = tmp_test_corpus.authors[0]

        df_delta = calculate_burrows_delta(tmp_train_corpus, tmp_test_corpus)

        ground_truths.extend(list(df_delta.index == true_author))
        delta_values.extend(list(df_delta.iloc[:, 0]))

    return np.asarray(ground_truths), np.asarray(delta_values)


def calibrate(corpus, model=LogisticRegression(class_weight="balanced")):
    ground_truths, delta_values = get_calibration_curve(corpus)

    model.fit(np.reshape(delta_values, (-1, 1)), ground_truths)

    corpus.probability_model = model


def predict_proba(train_corpus: Corpus, test_corpus: Corpus) -> pd.DataFrame:
    """
    Returns the probability that the test corpus is by the same author as the training corpus.

    :param train_corpus: The corpus to serve as a baseline for the word frequency calculations.
    :param test_corpus: The corpus to compare it with.
    :return: The probability according to the calibrated model, that the test corpus was by the same author as the train corpus.
    """
    df_delta = calculate_burrows_delta(train_corpus, test_corpus)

    df_probas = pd.DataFrame()
    for test_author_idx in range(df_delta.shape[1]):
        values = train_corpus.probability_model.predict_proba(
            np.reshape(list(df_delta.iloc[:, test_author_idx]), (-1, 1)))[:,
                 1]
        df_probas[df_delta.columns[test_author_idx]] = list(values)

    df_probas.index = df_delta.index

    return df_probas



import operator
from collections import Counter

import numpy as np
import pandas as pd

from faststylometry import Corpus


def get_top_tokens(corpus: Corpus, vocab_size: int) -> list:
    """
    Identify the n highest ranking tokens in the corpus.

    :param corpus: The corpus to look for the commonest words in. The corpus must be tokenised.
    :param vocab_size: The n most common words in the corpus will be used.
    :return: A list of the n most common tokens in the corpus.
    """

    assert (len(corpus.tokens) > 0)  # you should have tokenised the corpus.

    token_freqs = Counter()
    for token_seq in corpus.tokens:
        for token in token_seq:
            token_freqs[token] += 1

    corpus.top_tokens = [tok for tok, freq in
                         sorted(token_freqs.items(), key=operator.itemgetter(1), reverse=True)[:vocab_size]]
    corpus.top_tokens_set = set(corpus.top_tokens)

    return corpus.top_tokens


def set_top_tokens(corpus: Corpus, top_tokens: list):
    """
    Use a predefined set of tokens to calculate Burrows' Delta on this corpus.
    You would do this, if the corpus is a test corpus, and your model has been trained on a training corpus.

    :param corpus: the corpus you want to set the top tokens for.
    :param top_tokens: the top tokens which you have calculated elsewhere, e.g. from a training corpus.
    """
    corpus.top_tokens = top_tokens
    corpus.top_tokens_set = set(top_tokens)


def get_token_counts(corpus: Corpus):
    """
    Calculate the number of times each of the top tokens occurs in the corpus, and store this data in two dataframes in the corpus.

    :param corpus: The corpus to run the operation on and store the result in.
    """
    token_counts = np.zeros((len(corpus.authors), len(corpus.top_tokens)))

    total_token_counts = np.asarray([len(tokens) for tokens in corpus.tokens])

    for idx in range(len(corpus.authors)):

        this_work_token_counts = Counter()
        for tok in corpus.tokens[idx]:
            if tok in corpus.top_tokens_set:
                this_work_token_counts[tok] += 1

        for tok_idx, tok in enumerate(corpus.top_tokens):
            token_counts[idx, tok_idx] = this_work_token_counts.get(tok, 0)

    corpus.df_token_counts = pd.DataFrame(token_counts, columns=corpus.top_tokens)

    corpus.df_token_counts["author"] = corpus.authors
    author_book_combinations = [a + " - " + b for a, b in zip(corpus.authors, corpus.books)]
    corpus.df_token_counts["author_book"] = author_book_combinations

    corpus.df_total_token_counts = pd.DataFrame({"count": total_token_counts})
    corpus.df_total_token_counts["author"] = corpus.authors
    corpus.df_total_token_counts["author_book"] = author_book_combinations


def get_token_counts_by_author(corpus: Corpus):
    """
    Group the token counts across works by the same author, so we have a token count for each author.
    :param corpus:  The corpus to run the operation on and store the result in.
    """
    corpus.df_token_counts_by_author = corpus.df_token_counts.groupby("author").sum()
    corpus.df_total_token_counts_by_author = corpus.df_total_token_counts.groupby("author").sum()


def get_token_counts_by_author_and_book(corpus: Corpus):
    """
    Group the token counts across works by the same author and book, so we have a token count for each book.
    :param corpus:  The corpus to run the operation on and store the result in.
    """
    corpus.df_token_counts_by_author = corpus.df_token_counts.groupby("author_book").sum()
    corpus.df_total_token_counts_by_author = corpus.df_total_token_counts.groupby("author_book").sum()


def get_token_proportions(corpus: Corpus):
    """
    Calculate the fraction of words in each author's subcorpus which are equal to each word in our vocabulary.
    :param corpus:  The corpus to run the operation on and store the result in.
    """
    token_proportions = corpus.df_token_counts_by_author.to_numpy() / corpus.df_total_token_counts_by_author.to_numpy()

    corpus.df_token_proportions = pd.DataFrame(token_proportions, columns=corpus.top_tokens,
                                               index=corpus.df_token_counts_by_author.index)


def get_author_z_scores(test_corpus: Corpus, training_corpus: Corpus = None) -> pd.DataFrame:
    """
    Calculate the Z-score relating the test corpus to each author's subcorpus in the training corpus.
    :param test_corpus:  The corpus to run the operation on and store the result in.
    :param training_corpus: The training corpus with a number of authors' subcorpora, which will be compared to the test corpus.
    :return: A dataframe of Z-scores for each author in the training corpus.
    """
    if not training_corpus:
        training_corpus = test_corpus
    test_corpus.df_author_z_scores = (
                                             test_corpus.df_token_proportions - training_corpus.df_token_proportions.mean()) / training_corpus.df_token_proportions.std()

    return test_corpus.df_author_z_scores


def calculate_difference_from_train_corpus(test_corpus, train_corpus=None):
    """
    Calculate the difference between the Z-scores of the test corpus and the Z-scores of each author in the training corpus.
    :param test_corpus:  The corpus to run the operation on and store the result in.
    :param train_corpus: The training corpus with a number of authors' subcorpora, which will be compared to the test corpus.
    """
    test_corpus.df_difference = [None] * len(test_corpus.df_author_z_scores)
    for i in range(len(test_corpus.df_author_z_scores)):
        test_corpus.df_difference[i] = train_corpus.df_author_z_scores.sub(test_corpus.df_author_z_scores.iloc[i, :])


def calculate_burrows_delta(train_corpus: Corpus, test_corpus: Corpus, vocab_size: int = 50) -> pd.DataFrame:
    """
    Calculate the Burrows' Delta statistic for the test corpus vs every author's subcorpus in the training corpus.
    :param train_corpus: A corpus of known authors, which we will use as a benchmark to compare to the test corpus by an unknown author.
    :param test_corpus: The corpus by an unknown author.
    :param vocab_size: We will take the top n tokens from the training corpus and use as the vocabulary for the model. Normally 50-100 make sensible values.
    :return: A DataFrame of Burrows' Delta values for each author in the training corpus.
    """
    get_top_tokens(train_corpus, vocab_size)
    get_token_counts(train_corpus)
    get_token_counts_by_author(train_corpus)
    get_token_proportions(train_corpus)
    get_author_z_scores(train_corpus)

    set_top_tokens(test_corpus, train_corpus.top_tokens)
    get_token_counts(test_corpus)
    get_token_counts_by_author_and_book(test_corpus)
    get_token_proportions(test_corpus)
    get_author_z_scores(test_corpus, train_corpus)

    calculate_difference_from_train_corpus(test_corpus, train_corpus)

    deltas = {}
    for test_author_idx in range(len(test_corpus.df_author_z_scores)):
        deltas[test_corpus.df_author_z_scores.index[test_author_idx]] = test_corpus.df_difference[
            test_author_idx].abs().mean(axis=1)

    df_delta = pd.concat(deltas, axis=1)

    return df_delta



from faststylometry.corpus import Corpus
from faststylometry.util import load_corpus_from_folder
from faststylometry.en import tokenise_remove_pronouns_en
from faststylometry.burrows_delta import calculate_burrows_delta
from faststylometry.probability import predict_proba, calibrate


import re

import nltk

is_alpha_pattern = re.compile(r'^\w+$')

# pronouns
stopwords = {"he",
             "her",
             "hers",
             "herself",
             "him",
             "himself",
             "his",
             "i",
             "me",
             "mine",
             "my",
             "myself",
             "our",
             "ours",
             "ourselves",
             "she",
             "thee",
             "their",
             "them",
             "themselves",
             "they",
             "thou",
             "thy",
             "thyself",
             "us",
             "we",
             "ye",
             "you",
             "your",
             "yours",
             "yourself", }


def tokenise_remove_pronouns_en(text: str) -> list:
    """
    Tokenise a sentence according to Englsh rules, and remove all pronouns.

    :param text: the original sentence.
    :return: all non-pronoun tokens.
    """
    tokens = [tok for tok in nltk.word_tokenize(text.lower()) if is_alpha_pattern.match(tok)]

    tokens_without_stopwords = [tok for tok in tokens if tok not in stopwords]

    return tokens_without_stopwords



import os
import re

from faststylometry.corpus import Corpus


def load_corpus_from_folder(folder: str, pattern: str = None) -> Corpus:
    """
    Iterates over the .txt files in a folder and adds their contents to the corpus.

    Assumes that files are of the format [author]_-_[title].txt.

    :param folder: the folder ony our local system to iterate over.
    :param pattern: an optional filter to apply to the filenames.
    :return: a corpus of the files found.
    """
    corpus = Corpus()
    for root, _, files in os.walk(folder):
        for filename in files:
            if pattern is not None and not pattern in filename:
                continue
            if filename.endswith(".txt") and "_" in filename:
                with open(os.path.join(root, filename), "r", encoding="utf-8") as f:
                    text = f.read()
                author, book = re.split("_-_", re.sub(r'\.txt', '', filename))

                corpus.add_book(author, book, text)
    return corpus



class Corpus:

    def __init__(self, authors: list = None, books: list = None, tokens: list = None):
        """
        Create a new corpus object, which is empty by default, but which can be initialised optionally with a set of books.
        :param authors: A list of strings representing author names. If a set of books are being used to initialise the corpus, len(authors) must equal len(books) and len(tokens)
        :param books:  A list of strings representing book titles.
        :param tokens: A list of lists representing tokens in the books.
        """
        if authors and books and tokens:
            self.authors = authors
            self.books = books
            self.tokens = tokens
            self.texts = None
        else:
            self.authors = []
            self.books = []
            self.tokens = []
            self.texts = []

    def add_book(self, author: str, book: str, text: str):
        """
        Add a single book to the corpus. This can only be done if the corpus has not been initialised with books in the constructor.
        :param author: The author's name as string.
        :param book: The book title as string.
        :param text: The book content as string.
        """
        assert (len(self.tokens) == 0)
        self.authors.append(author)
        self.books.append(book)
        self.texts.append(text)

    def tokenise(self, tokenise):
        """
        Tokenise all books in the corpus using the custom tokenisation function.

        :param tokenise: a tokenise function which takes a str and returns a list of tokens. It is language-specific and should remove pronouns.
        """
        self.tokens = [tokenise(text) for text in self.texts]

    def split(self, segment_length: int = 1000):
        """
        Split the books in the corpus into smaller books of a specified maximum length. So if segment_length = 1000 and you have a book of length 1500,
        it will be split into two books of lengths 1000 and 500 respectively, with the same author and title attributes.
        This Corpus object is not modified but a new instance will be constructed.

        :param segment_length: The maximum length of books in the new corpus.
        :return: The new corpus with split books.
        """
        new_authors = []
        new_books = []
        new_tokens = []

        for author_id in range(len(self.authors)):
            for ctr, segment_start in enumerate(range(0, len(self.tokens[author_id]), segment_length)):
                segment_tokens = self.tokens[author_id][
                                 segment_start:min(segment_start + segment_length, len(self.tokens[author_id]))]
                new_authors.append(self.authors[author_id])
                new_books.append(self.books[author_id] + "_" + str(ctr))
                new_tokens.append(segment_tokens)

        return Corpus(new_authors, new_books, new_tokens)



from distutils.core import setup

with open("README.rst", "r") as fh:
  long_description = fh.read()

setup(
  name = 'faststylometry',
  packages = ['faststylometry'],
  version = '0.3',
  license='MIT', 
  description = 'Calculates Burrows Delta',
  long_description=long_description,
  author = 'Thomas Wood',
  #author_email = 'thomas@fastdatascience.com',
  url = 'https://freelancedatascientist.net',
  keywords = ['stylometry', 'nlp', 'burrows delta', 'delta', 'forensic stylometry', 'natural language processing'],
  install_requires=[
          'numpy>=1.18.5',
          'pandas>=1.1.2',
          'scikit-learn>=0.23.1'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',   
    'Intended Audience :: Developers',   
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',   
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
  ],
  include_package_data=True,
)
